// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef GRIDPOINT_HPP
#define GRIDPOINT_HPP

#include <stdint.h>
#include "DlxHead.hpp"
#include "Point.hpp"
#include "makeArray.hpp"
#include "util.hpp"

// For FILA the occupancy state of the puzzle is tracked in a bit field.  The
// size of the bit field is determined via the preprocessor definition
// GRIDBITFIELD_SIZE which may be one of 32, 64, or 128.  If not set at the
// compile line, it defaults to 64 here.  The size of the bit field puts a
// limit on the maximum size of the untiled region before these algorithms can
// be activated.  But using the maximum size of 128 probably slows the program
// for 64 bit hardware architectures.  (I have not yet done a performance test
// as I write this.)
//
#ifndef GRIDBITFIELD_SIZE
#define GRIDBITFIELD_SIZE 64
#endif

#if GRIDBITFIELD_SIZE == 32
#define GRIDBITFIELD_T uint32_t
#elif GRIDBITFIELD_SIZE == 64
#define GRIDBITFIELD_T uint64_t
#elif GRIDBITFIELD_SIZE == 128
#define GRIDBITFIELD_T __uint128_t
#else
#error Preprocessor definition of GRIDBITFIELD_SIZE must be one of 32, 64, or 128.
#endif

typedef GRIDBITFIELD_T gridbitfield_t;
const int maxFilaGridPoints = 8*sizeof(gridbitfield_t);

/** An important model element of the Puzzle is the GridPoint which represents
 ** a hole in the puzzle.  GridPoints are not instantiated at puzzle locations
 ** occupied by stationary pieces.  ** GridPoints are accessible in a number of ways:
 **
 ** 1. Given the (x,y,z) coordinate, you can access the GridPoint at that
 **    coordinate via *Puzzle.grid[x][y][z].  Note that Puzzle.grid[x][y][z]
 **    is NULL for those puzzle locations occupied by stationary pieces.
 **
 ** 2. All GridPoints are initially allocated as a single contiguous array
 **    whose location in memory is given by Puzzle.gridStore.
 **
 ** 3. A GridPoint is a DlxHead object and so the list of currently unoccupied
 **    GridPoints is maintained by the Dlx search algorithm.
 **
 ** 4. Each Image has a layout member which is a vector of pointers to the
 **    GridPoints it occupies.
 **/
class GridPoint : public Point, public DlxHead
{
    public:
        /** All GridPoints are given a fixed sequential 0-based index in an
         ** (x, y, z) sorting priority.  This id and has no direct
         ** relationship to bit.
         **/
        int                        id;

        /** A bit field with a single bit turned on.  This field is only set
         ** and used if Tiling is enabled.  If the number of GridPoints in the
         ** puzzle does not exceed the size of the field, then this bit
         ** property is set up front and is permanent.  But if the size of the
         ** puzzle is too large to be modeled by gridbitfield_t, then bit field
         ** must be reassigned each time the number of remaining holes in the
         ** puzzle is reduced to the number of bits in gridbitfield_t.  These
         ** bits in turn are used to generate the layoutMask for each Image
         ** (i.e., each Dlx row) that remains in the DLX matrix.
         **/
        gridbitfield_t             bit;

        /** bitNum = log_2(bit)
         **/
        int                        bitNum;

        /** This field is used to verify that user-defined stationary pieces
         ** don't collide and also to support volume checks.  Any negative
         ** value for the fill field means the GridPoint is unoccupied.  The
         ** exact negative value used to indicate an unoccupied GridPoint is
         ** either -1 or -2 and is given by the unoccupiedFill variable.  (I
         ** explain why it's done this way shortly -- bear with me.)
         ** Initially, unoccupiedFill is set to -1 and accordingly the fill
         ** field for all GridPoints is initialized to -1.  When an image is
         ** pushed to the imageStack, the fill field for the GridPoints that
         ** image occupies is updated to the index of that image in the stack.
         ** When an image is popped from the stack, the fill field for those
         ** same GridPoints is set to the current value of unoccupiedFill.
         ** Before a stationary piece is pushed to the stack, it is verified
         ** that it only occupies GridPoints that are not yet filled.  Volume
         ** checks are performed through a simple fill algorithm that changes
         ** the value of unoccupiedFill from -1 to -2 (or from -2 to -1) and
         ** measures the volume of subspaces as it goes.  Upon completion all
         ** unoccupied GridPoints will have changed state from -1 to -2 (or
         ** vice-versa), and the unoccupiedFill setting is updated
         ** accordingly.  (So by allowing either -1 or -2 to represent the
         ** unoccupied state, the fill algorithm doesn't have to revert the
         ** state of the fill field for unoccupied GridPoints to their
         ** previous setting improving the performance of volume checks.  The
         ** fill updates required with each piece placement and unplacement is
         ** expensive, but all such fill processing is disabled once the
         ** numRemainingPieces falls below both the volumeCheck and
         ** volumePrune thresholds.
         **/
        int                        fill;

        /** This field is initialized to the binary-or of each bit of each of
         ** the six neighbors in each ordinal direction (or only the three
         ** neighbors in positive ordinal direction if only the F heuristic is
         ** used with FILA).  If there is no neighbor in a direction, then
         ** (obviously) no bit is set for that (non-existent) neighbor.  Note
         ** that bits are only assigned (and reassigned) each time DLX is
         ** deactivated and FILA is activated.  So not all neighbors are
         ** included -- only those that are still empty the last time DLX was
         ** deactivated.
         **
         ** The neighborMask is only used by the E OrderingHeuristic, which
         ** ands the neighborMask of a GridPoint with the complement of the
         ** puzzle occupancy state to get a bit field representing the
         ** open neighbors of this GridPoint.  The number of bits set
         ** in this bit field then gives a count of the number of open
         ** neighbors, which is used as a measure of how constrained
         ** the hole is (and therefore a relative estimate of how many
         ** images are likely to fill that hole.)
         **/
        gridbitfield_t             neighborMask;

        /** These bit fields identify the bits for the six neighbors of this
         ** GridPoint.  Remember that bits are only defined once the puzzle is
         ** converted from the DLX data model to the FILA data model.  If a
         ** neighbor is non-existent (either not part of the puzzle or already
         ** filled when FILA is enabled), the value is 0.  They are only used
         ** when NOF is enabled.  If NOF is enabled, they are used by
         ** OrderingHeuristics to generate ImageListSet indicies (ilsi).
         **/
        gridbitfield_t             nbr_x; // grid[x+1][y][z].bit or 0
        gridbitfield_t             nbr_y; // grid[x][y+1][z].bit or 0
        gridbitfield_t             nbr_z; // grid[x][y][z+1].bit or 0
        gridbitfield_t             nbr_X; // grid[x-1][y][z].bit or 0
        gridbitfield_t             nbr_Y; // grid[x][y-1][z].bit or 0
        gridbitfield_t             nbr_Z; // grid[x][y][z-1].bit or 0

        /** List of neighboring GridPoints.  This list is only initialized for
         ** GridPoints that are NOT filled by stationary pieces.  This list is
         ** currently only used by Puzzle::fillVolume().  This routine doesn't
         ** need to know about neighbors that are part of stationary pieces.
         ** Accordingly, only those neighbors that are not filled by
         ** stationary pieces is included in the list.
         **/
        std::vector<GridPoint*>     neighborList;

        GridPoint();
        virtual ~GridPoint();
        bool isGridPoint() const override;
        std::ostream& print(std::ostream& os) const override;
};


inline GridPoint::GridPoint()
    :   id(-1),
        bit(0),
        fill(-1),
        neighborMask(0),
        nbr_x(0),
        nbr_y(0),
        nbr_z(0),
        nbr_X(0),
        nbr_Y(0),
        nbr_Z(0)
{
}

inline GridPoint::~GridPoint()
{
}


inline bool GridPoint::isGridPoint() const
{
    return true;
}


inline std::ostream& GridPoint::print(std::ostream& os) const
{
    Point::print(os) << ":";
    DlxHead::print(os)
        << ":id=" << id
        << ":bitNum=" << bitNum
        //<< ":bit=" << Bits<gridbitfield_t>(bit)
        //<< ":neighborMask=" << Bits<gridbitfield_t>(neighborMask)
        //<< ":nbr_x=" << Bits<gridbitfield_t>(nbr_x)
        //<< ":nbr_y=" << Bits<gridbitfield_t>(nbr_y)
        //<< ":nbr_z=" << Bits<gridbitfield_t>(nbr_z)
        //<< ":nbr_X=" << Bits<gridbitfield_t>(nbr_X)
        //<< ":nbr_Y=" << Bits<gridbitfield_t>(nbr_Y)
        //<< ":nbr_Z=" << Bits<gridbitfield_t>(nbr_Z)
        ;
    return os;
}


// Seems like there should be some way to avoid writing this.
//
inline std::ostream& operator << (std::ostream& os, const GridPoint& gp)
{
    return gp.print(os);
}

#endif
